#!/usr/bin/env python

"""Contains functionality that doesn't fit elsewhere
"""

import collections.abc
import inspect
import functools

class switch(object):
    """Simulate a switch-case statement.

    http://code.activestate.com/recipes/410692/
    """

    def __init__(self, value):
        self.value = value
        self.fall = False

    def __iter__(self):
        """Return the match method once, then stop"""
        yield self.match
        raise StopIteration
    
    def match(self, *args):
        """Indicate whether or not to enter a case suite"""
        if self.fall or not args:
            return True
        elif self.value in args:
            self.fall = True
            return True
        else:
            return False

# Following inspired by http://stackoverflow.com/a/7811344/974555
def validate(func, locals):
    """Validate function with arguments.

    Inside an annotated function (see PEP-3107), do type checking on the
    arguments.  An annotation may be either a type or a callable.  Use
    like this:

    def f(x: str, b: int):
        validate(f, locals())
        ... # proceed

    or use the validator annotation:
    @validator
    def f(x: str, b: int):
        ... # proceed
    """
    for var, test in func.__annotations__.items():
        value = locals[var]
        _validate_one(var, test, value)

def _validate_one(var, test, value):
    """Verify that var=value passes test

    Internal function for validate
    """
    if isinstance(test, type): # check for subclass
        if not isinstance(value, test):
            raise TypeError(("Wrong type for argument '{}'.  "
                   "Expected: {}.  Got: {}.").format(
                        var, test, type(value)))
    elif callable(test):
        if not test(value):
            raise TypeError(("Failed test for argument '{0}'.  "
                             "Value: {1}.  Test {2.__name__} "
                             "failed.").format(
                var, value if len(repr(value)) < 10000 else "(too long)", test))
    elif isinstance(test, collections.abc.Sequence): # at least one should be true
        passed = False
        for t in test:
            try:
                _validate_one(var, t, value)
            except TypeError:
                pass
            else:
                passed = True
        if not passed:
            raise TypeError(("All tests failed for argument '{0}'. "
                             "Value: {1}.  Tests: {2!s}").format(
                var, value, test))
    else:
        raise RuntimeError("I don't know how to validate test {}!".format(test))

def validator(func):
    """Decorator to automagically validate a function with arguments.

    Uses functionality in 'validate', required types/values determined
    from decorations.  Example::

        @validator
        def f(x: numbers.Number, y: numbers.Number, mode: str):
            return x+y

    Does not currently work for *args and **kwargs style arguments.
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        fsig = inspect.signature(func)
        # initialise with defaults; update with positional arguments; then
        # update with keyword arguments
        lcls = dict([(x.name, x.default) for x in fsig.parameters.values()])
        lcls.update(**dict(zip([x.name for x in fsig.parameters.values()], args)))
        lcls.update(**kwargs)
        validate(func, lcls)
        return func(*args, **kwargs)

    return inner

def cat(*args):
    """Concatenate either ndarray or ma.MaskedArray

    Arguments as for numpy.concatenate or numpy.ma.concatenate.
    First argument determines type.
    """

    if isinstance(args[0][0], numpy.ma.MaskedArray):
        return numpy.ma.concatenate(*args)
    else:
        return numpy.concatenate(*args)



"""Contains classes and functionally to calculate collocations
"""
# pylint: disable-msg=E1101

import math
import numpy
import numpy.ma
import logging

import pyproj

from . import dataset
from . import stats
from . import geo
from . import tools

class CollocatedDataset(dataset.HomemadeDataset):
    """Holds collocations.

    Attributes:

    primary
    secondary
    max_distance    Maximum distance in m
    max_interval    Maximum time interval in s.
    projection      projection to use in calculations

    The following attributes may be changed at your own risk.  Changing
    should not affect results, but may affect performance.  Optimise based
    on application.  Subject to change.

    bin_interval_time
    bin_interval_lat
    bin_interval_lon
    """

    primary = None
    secondary = None
    projection = "WGS84"
    ellipsoid = None

    max_distance = 0.0 # distance in m

    _max_interval = numpy.timedelta64(0, 's')
    @property
    def max_interval(self):
        """Maximum interval time.

        Can be set as a number, interpreted in seconds, or as a
        timedelta64 object.
        """
        return self._max_interval

    @max_interval.setter
    def max_interval(self, value):
        self._max_interval = numpy.timedelta64(value, 's')

    bin_interval_time = numpy.timedelta64(1, 'D')
    bin_interval_lat = 1.0 # degree
    bin_interval_lon = 1.0 # degree

    def __init__(self, primary, secondary, **kwargs):
        """Initialize with Dataset objects
        """
        self.primary = primary
        self.secondary = secondary
        if "projection" in kwargs:
            self.projection = kwargs.pop("projection")
        self.ellipsoid = pyproj.Geod(ellps=self.projection)
        self.max_interval = 0

        super().__init__(**kwargs)

    def find_granule_pairs(self, start_date=None, end_date=None):
        """Iterate through all (prim, sec) co-time granule pairs

        Can optionally pass in start_date and end_date, that will be passed
        on to the primary find_granules.
        """

        if start_date is None:
            start_date = max([self.primary.start_date,
                self.secondary.start_date])

        if end_date is None:
            end_date = min([self.primary.end_date,
                self.secondary.end_date])

        for gran_prim in self.primary.find_granules_sorted(start_date, end_date):
            for gran_sec in self.secondary.find_granules_sorted(
                *self.primary.get_times_for_granule(gran_prim)):
                yield (gran_prim, gran_sec)

    def read_aggregated_pairs(self, start_date=None, end_date=None):
        """Iterate and read aggregated co-time granule pairs

        Collect and read all secondary granules sharing the same primary.
        """
        old_prim = old_sec = None
        prim = None
        sec = []
        logging.info("Searching granule pairs")
        for (gran_prim, gran_sec) in self.find_granule_pairs(
            start_date, end_date):
#            logging.debug("Next pair: {} vs. {}".format(
#                gran_prim, gran_sec))
            if gran_prim != old_prim: # new primary
                logging.debug("New primary: {!s}".format(gran_prim))
                if prim is not None: # if not first time, yield pair
                    logging.debug("Yielding!")
                    yield (prim, numpy.concatenate(sec))
                prim = self.primary.read(gran_prim)
                old_prim = gran_prim

            if gran_sec != old_sec:
                logging.debug("New secondary: {!s}".format(gran_sec))
                try:
                    sec.append(self.secondary.read(gran_sec))
                except ValueError as msg:
                    logging.error("Could not read {}: {}".format(
                        gran_sec, msg.args[0]))
                    continue
                old_sec = gran_sec

        yield (prim, numpy.concatenate(sec))

                

    def collocate_period(self, start_date=None, end_date=None):
        """Collocate period from start_date to end_date

        And return results
        """
        all_p_col = []
        all_s_col = []
#        all_p_ind = []

        # construct dtype
#        L = [[("i"+s, "i4"), ("lat"+s, "f8"), ("lon"+s, "f8"),
#            ("time"+s, "M8[s]")] for s in "12"]
        
#        dtp = L[0] + L[1]

        # FIXME: aggregate multiple granules before collocating, in
        # particular for cases with only one measurement per file
#        for (gran_prim, gran_sec) in self.find_granule_pairs(
#                start_date, end_date):
#            logging.info("Collocating {0!s} with {1!s}".format(
#                ran_prim, gran_sec))
#            prim = self.primary.read(gran_prim)
#            sec = self.secondary.read(gran_sec)
        logging.info(("Searching collocations {!s} vs. {!s}, distance {:.1f} km, "
                      "interval {!s}, period {!s} - {!s}").format(
                        self.primary.name, self.secondary.name,
                        self.max_distance/1e3, self.max_interval,
                        start_date, end_date))
        for (prim, sec) in self.read_aggregated_pairs(
                start_date, end_date):
            logging.info(("{} {:d} measurements spanning {!s} - {!s}, "
                          "{} {:d} measurements spanning {!s} - {!s}").format(
                    self.primary.name, prim.shape[0], min(prim["time"]),
                    max(prim["time"]),
                    self.secondary.name, sec.shape[0], min(sec["time"]),
                    max(sec["time"])))
            #(p_col, s_col) = self.collocate(prim, sec)
            logging.info("Collocating...")
            p_ind = self.collocate(prim, sec)
            logging.info("Found {0:d} collocations".format(p_ind.shape[0]))
            #all_p_ind.append(p_ind) # FIXME: get more useful than indices
            all_p_col.append(prim[p_ind[:, 0]])
            all_s_col.append(sec[p_ind[:, 1]])
            #all_s_col.append(s_col)
        #return numpy.concatenate(all_p_ind)
        return (numpy.concatenate(all_p_col), numpy.concatenate(all_s_col))

    def collocate_all(self, distance=0, interval=numpy.timedelta64(1, 's')):
        """Collocate all available data.
        """
        raise NotImplementedError("Not implemented yet")

    @tools.validator
    def collocate(self, arr1:geo.valid_geo, arr2:geo.valid_geo):
        """Collocate arrays in time, late, lon.

        Each of `arr1` and `arr2` must have ["time"] (datetime64),
        ["lat"] (float), and ["lon"] (float).

        Note that this is a low-level function, and you will likely want
        to call a higher level method such as collocate_all.

        :returns: N x 2 array with primary and secondary indices
        """

        # This algorithm can be optimised in a number of different ways:
        #
        # - Use quadtrees on a spherical grid instead of guessing grid
        # sizes
        # - Process only time that is in common
        # - For memory, loop through time and collocate bit by bit

        if self.max_interval == 0 or self.max_distance == 0:
            return numpy.empty(shape=(0, 2), dtype=numpy.uint64)

        # all binning should be by truncation, not rounding; i.e.
        # 2010-01-01 23:00:00 is binned on 2010-01-01.

        # first bin both by time, which is a special case because
        # numpy.digitize, on which pyatmlab.stats.bin_nd relies, does not
        # support it; so we need to truncate both time series to a common
        # format, then use ints for the binning

        # FIXME: this can be optimized by doing only further processing
        # for common time interval

        if (arr1["time"].max() + self.max_interval < arr2["time"].min() or
            arr2["time"].max() + self.max_interval < arr1["time"].min()):
            return numpy.empty(shape=(0, 2), dtype=numpy.uint64)

        # will finally want to return indices rather than copying actual
        # data, keep track of those
        ind1 = numpy.arange(arr1.shape[0])
        ind2 = numpy.arange(arr2.shape[0])

        # truncate time series to resolution of self.bin_interval_time
        newtype = "<M8[{}]".format(self.bin_interval_time.dtype.str[-2])
        times_trunc = [arr["time"].astype(newtype) for arr in (arr1, arr2)]
        times_int = [time.astype(numpy.int64) for time in times_trunc]
#        time_bins = numpy.arange(
#            min(t.min() for t in times_trunc),
#            max(t.max() for t in times_trunc),
#            self.bin_interval_time)

        lats = [arr1["lat"], arr2["lat"]]
        lons = [arr1["lon"], arr2["lon"]]

        bins = {}
        bin_intervals = dict(lat=self.bin_interval_lat,
            lon=self.bin_interval_lon, time=self.bin_interval_time)
        for k in ("lat", "lon"):
            kmax = max(v.max() for v in (arr1[k], arr2[k]))
            kmin = min(v.min() for v in (arr1[k], arr2[k]))
            bins[k] = numpy.linspace(
                kmin-1, kmax+1,
                max((((kmax+1)-(kmin-1))/bin_intervals[k], 2)))

        tmin = min(t.min() for t in times_trunc).astype(numpy.int64)
        tmax = max(t.max() for t in times_trunc).astype(numpy.int64)
        bins["time"] = numpy.linspace(
            tmin-1, tmax+1, ((tmax+1)-(tmin-1)) /
                bin_intervals["time"].astype(numpy.int64)).astype(newtype)
                    
#        lat_bins = numpy.linspace(
#            numpy.floor(min(lat.min() for lat in lats)-1),
#            numpy.ceil(max(lat.max() for lat in lats)+1),
#        lat_bins = numpy.arange(
#            numpy.floor(min(lat.min() for lat in lats)),
#            numpy.ceil(max(lat.max() for lat in lats)+1),
#            self.bin_interval_lat)

        # Note: this will be too large if longitudes cross (anti)meridian
#        lon_bins = numpy.arange(
#            numpy.floor(min(lon.min() for lon in lons)),
#            numpy.ceil(max(lon.max() for lon in lons)+1),
#            self.bin_interval_lon)

        # Perform the actual binning for primary, secondary.  This part
        # could be optimised a lot, ideally using quadtrees or at least by
        # guessing a more appropriate grid size.
        binned = [stats.bin_nd(
            [times_int[i], lats[i], lons[i]],
            [bins["time"].astype(numpy.int64), bins["lat"], bins["lon"]])
            for i in (0, 1)]

        # count the number of entries per bin
        bin_no = numpy.array([numpy.array([b.size for b in bb.flat]).reshape(bb.shape)
                for bb in binned])

        # some intermediate checking to verify all got binned
        if bin_no[0, ...].sum() != arr1.size or bin_no[1, ...].sum() != arr2.size:
            raise RuntimeError("Some data remained unbinned!")

        # number of neighbouring bins to look into
        binrange_time = math.ceil(self.max_interval/self.bin_interval_time)
        cell_height = 2 * math.pi * self.ellipsoid.b / 360
        cell_width = (2 * math.pi * numpy.cos(numpy.deg2rad(bins["lat"])) *
                      self.ellipsoid.b / 360)

        binrange_lat = numpy.ceil(self.max_distance/
            (self.bin_interval_lat*cell_height))
        binrange_lon = numpy.ceil(self.max_distance/
            (self.bin_interval_lon*cell_width))

        #all_p_met = []
        #all_s_met = []
        all_ind_met = []

        for time_i in range(len(bins["time"])):
            # range of secondary time bins
            t_s_min = max(0, time_i - binrange_time)
            t_s_max = min(bins["time"].size, time_i + binrange_time + 1)

            # potentially skip lat & lon loops
            if (bin_no[0, time_i, :, :].max() == 0 or
                    bin_no[1, t_s_min:t_s_max, :, :].max() == 0):
                continue

            for lat_i in range(len(bins["lat"])):
                # range of secondary lat bins
                lat_s_min = max(0, lat_i - binrange_lat)
                lat_s_max = min(bins["lat"].size, lat_i + binrange_lat + 1)

                # potentially skip lon loop
                if (bin_no[0, time_i, lat_i, :].max() == 0 or
                        bin_no[1, t_s_min:t_s_max, lat_s_min:lat_s_max, :].max() == 0):
                    continue

                max_lon_range = max(binrange_lon[lat_s_min:lat_s_max])
                for lon_i in range(len(bins["lon"])):
                    # range of secondary lon bins

                    # for width of lons consider polemost relevant
                    # latitude bin
                    lon_is = numpy.mod(numpy.arange(lon_i - max_lon_range,
                            lon_i+max_lon_range), bins["lon"].size).astype('uint64')
                    #lon_s_min = max(0, lon_i - max_lon_range)
                    #lon_s_max = min(lon_bins.size-1, lon_i + max_lon_range + 1)

                    if (bin_no[0, time_i, lat_i, lon_i].max() == 0 or
                            bin_no[1, t_s_min:t_s_max,
                            lat_s_min:lat_s_max,
                            lon_is].sum() == 0):
                        continue

                    selec1 = binned[0][time_i, lat_i, lon_i]
                    selec2 = numpy.ma.concatenate(binned[1][
                        t_s_min:t_s_max,
                        lat_s_min:lat_s_max,
                        lon_is].ravel().tolist())

                    primary = arr1[selec1]
                    secondary = arr2[selec2]

                    prim_ind = ind1[selec1]
                    sec_ind = ind2[selec2]

                    ind_met = self._collocate_bucket(
                        primary, secondary, prim_ind, sec_ind)

                    all_ind_met.append(ind_met)
                    #all_p_met.append(p_met)
                    #all_s_met.append(s_met)

        if len(all_ind_met) > 0:
            return numpy.concatenate(all_ind_met)
        else:
            return numpy.empty(shape=(0, 2), dtype=numpy.uint64)
            
#        return tuple((numpy.ma.concatenate 
#            if isinstance(x, numpy.ma.MaskedArray)
#            else numpy.concatenate)(x)
#                for x in (all_p_met, all_s_met))

    def _collocate_bucket(self, primary, secondary, prim_ind, sec_ind):
        """Collocate a single bucket.  Internal function used by
        collocate.

        Expects two buckets containing measurements that will be
        brute-forced against each other, as well as corresponding indices.

        Returns an N x 2 ndarray with indices selected from prim_ind,
        sec_ind.
        """

        if primary.size == 0 or secondary.size == 0:
            return (numpy.empty(shape=(0,), dtype=numpy.int64),
                    numpy.empty(shape=(0,), dtype=numpy.int64))

        # find pairs meeting time criterion
        intervals = (primary[:, numpy.newaxis]["time"] -
                     secondary[numpy.newaxis, :]["time"])
        time_met = abs(intervals) < self.max_interval
        (time_met_i1, time_met_i2) = time_met.nonzero()

        # find pairs meeting distance criterion
        p_time_met = primary[time_met_i1]
        s_time_met = secondary[time_met_i2]

        p_ind_time_met = prim_ind[time_met_i1]
        s_ind_time_met = sec_ind[time_met_i2]

        (_, _, dist) = self.ellipsoid.inv(
            p_time_met["lon"], p_time_met["lat"],
            s_time_met["lon"], s_time_met["lat"],
            radians=False)

        dist_met = dist < self.max_distance
        #p_met = p_time_met[dist_met]
        p_ind_met = p_ind_time_met[dist_met]
        #s_met = s_time_met[dist_met]
        s_ind_met = s_ind_time_met[dist_met]

        #return p_met, s_met
        return numpy.array([p_ind_met, s_ind_met], dtype=numpy.uint64).T





#!/usr/bin/env python
# coding: utf-8

"""Various small mathematical functions

"""

import numpy
from . import tools
from .meta import expanddoc

inputs = """:param z: Height
    :type z: ndarray
    :param q: Quantity, dim 0 must be height.
    :type q: ndarray
    :param ignore_negative: Set negative values to 0
    :type ignore_negative: bool"""

@expanddoc
def layer2level(z, q, ignore_negative=False):
    """Converts layer to level. First dim. must be height.

    {inputs}
    :returns: Level-valued quantity.
    """
    dz = z[1:, ...] - z[:-1, ...]
    if ignore_negative:
        q[q<0]=0
    y_avg = (q[1:, ...] + q[:-1, ...])/2
    return (y_avg * dz)

@expanddoc
def integrate_with_height(z, q, ignore_negative=False):
    """Calculate vertically integrated value

    {inputs}
    :returns: Vertically integrated value
    """

    return layer2level(z, q, ignore_negative).sum(0)

@expanddoc
def cum_integrate_with_height(z, q, ignore_negative=False):
    """Calculate cumulative integral with height

    {inputs}
    :returns: Cumulatively vertically integrated value
    """
    return layer2level(z, q, ignore_negative).cumsum(0)

#@tools.validator # comment out because fails for *args
def average_position_sphere(*args: (lambda a: len(a) in (1,2))):
    """Calculate the average position for a set of angles on a sphere

    This is quite imprecise, errors can be dozens of km.  For more
    advanced calculations, use proj4/pyproj.

    Input can be either:

    :param lat: Vector of latitudes
    :param lon: Vector of longitudes

    Or:

    :param locs: Nx2 ndarray with lats in column 0, lons in column 1
    """

    if len(args) == 1:
        locs = args[0]
        lat = locs[0, :]
        lons = locs[1, :]
    elif len(args) == 2:
        lat, lon = args

    X = numpy.cos(lat) * numpy.cos(lon)
    Y = numpy.cos(lat) * numpy.sin(lon)
    Z = numpy.sin(lat)

    xm = X.mean()
    ym = Y.mean()
    zm = Z.mean()

    lonm = numpy.arctan2(ym, xm)
    latm = numpy.arctan2(zm, numpy.sqrt(xm**2+ym**2))

    return (latm, lonm)



"""Specific datasets
"""

from ..datasets import (TansoFTS, NDACCAmes, ACEFTS)
import pathlib
import datetime

datadir = pathlib.Path("/home/gerrit/sshfs/glacier/data/1/gholl/data")

tanso = TansoFTS(
    start_date = datetime.datetime(2010, 3, 23, 2, 24, 54, 210),
    end_date = datetime.datetime(2010, 10, 31, 20, 34, 50, 814),
    srcfile = (datadir / "1403050001" /
               "GOSATTFTS20100316_02P02TV0001R14030500010.h5"),
    name = "Tanso FTS")

ndacc_ames_eureka = NDACCAmes(basedir=datadir / "BrukerIFS",
    start_date = datetime.datetime(2006, 8, 1, 0, 0, 0),
    end_date = datetime.datetime(2010, 10, 1, 0, 0, 0),
    granule_cache_file = "times.dat",
    name = "NDACC Eureka")

acefts = ACEFTS(basedir=datadir / "ACE",
    start_date = datetime.datetime(2006, 1, 1, 0, 0, 0),
    end_date = datetime.datetime(2012, 10, 1, 0, 0, 0),
    granule_cache_file = "times.dat",
    name = "ACE FTS")



#!/usr/bin/env python

"""Describe batches of data: atmospheric db, scattering db, etc.

This module contains a class with functionality to read atmospheric data
from a variety of formats and write it to a variety of formats.

Mostly obtained from PyARTS.
"""

import copy

import numpy
import datetime
now = datetime.datetime.now

import numpy.lib.recfunctions
#import numpy.ma
import scipy.io

#import matplotlib.pyplot
#import matplotlib.cm

#from . import arts_math
from . import tools

class AtmosphericDatabase:
    """Represents an atmospheric database

    Apart from the default constructor, those constructors may be useful:

    - :func:`AtmosphericDatabase.from_evans07`
    - :func:`AtmosphericDatabase.from_evans12`

    Attributes:

    - ``data``
    - ``instrument``
    - ``name``
    """

    def __init__(self, **kwargs):
        """Constructor for atmospheric database

        Takes only keyword arguments, where each keyword ends up as an
        instance attribute, i.e.
        AtmosphericDatabase(instrument=pyatmlab.instruments.ici).
        """

        for (k, v) in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_evans07(cls, dbfile, instrument=None):
        """Read from db in style for Evans study in 2006-2007

        :param dbfile: Path to database-file
        :param instrument: Optional argument, pass instrument.
        """

        with open(dbfile, 'r') as fp:
            
            ncases = int(fp.readline().strip().split()[0])
            fp.readline() # ncloudpar ntemplay nrhlay nchan nztab nmutab ntinfo nuinfo
            fp.readline() # RH layers tops / bottoms
            chan_names = fp.readline().split("=")[0].strip().split()
            fp.readline() # ztab
            fp.readline() # mutab
            fp.readline() # header line

            # pre-allocate
            n_bt = len(chan_names)
            Z = numpy.zeros(ncases, dtype=cls.get_dtype(n_bt))

            i = 0
            while i<ncases:
                line = fp.readline()
                (iwp, dme, zmed, *rest) = (float(x) for x in line.strip().split())
                fp.readline() # empty line
                bt = numpy.array([float(x) for x in fp.readline().strip().split()])
                Z["BT"][i] = bt
                Z["Dme"][i] = dme
                Z["Zmed"][i] = zmed
                Z["IWP"][i] = iwp
                i += 1
                if i%(ncases//5) == 0:
                    print(now(), "done", "{}/{}".format(i, ncases))

        return cls(data=Z, instrument=instrument, name="colorado2007")

    @classmethod
    def from_evans12(cls, dbfile, instrument=None, profile=False,
                   radar=False, maxlen=None, ext=False):
        """Read from db in style for Evans study in 2012

        :param dbfile: Path to database-file
        :param instrument: Optional argument, pass instrument.
        :param profile: Include profile information.  Defaults to False.
        :param radar: Include radar information.  Defaults to False.
        :param int maxlen: Optional maximum length of db to read.
        :param bool ext: Read extended version, with manual additions by
            Gerrit 2014-01-16 (i.e. also outputting non-retrieved
            quantities used in the RTM)
        """

        print(now(), "Reading", dbfile)
        fp = open(dbfile, 'r')

        # skip and interpret header
        fp.readline() # header line
        ncases = int(fp.readline().strip().split()[0])
        maxlen = maxlen or ncases
        fp.readline() # no. of dimensions of independent Gaussian prior space
        fp.readline() # no. of elements in observation vector
        nchans = int(fp.readline().strip().split()[0])
        chan_names = fp.readline().split()
#            print(now(), "compare:", self.instrument.channel_string(), chan_names)

        # no. values per channel
        chan_lengths = numpy.array([int(c)
                                    for c in fp.readline().strip().split()])
        fp.readline() # no. of viewing angles
        fp.readline() # cos for each viewing angle
        fp.readline() # random seed
        fp.readline() # no. retrieved quantities
        fp.readline() # no. integrated cloud params
        n_layer_rh = int(fp.readline().strip().split()[0]) # no. retrieved humidity levels
        fp.readline() # height per humidity layer
        n_layer_ice = int(fp.readline().strip().split()[0]) # no. retrieved ice cloud layers
        (ice_bottom, ice_top) = (float(c) for c in
                                 fp.readline().split('!')[0].strip().split()) # bottom and top height ice cloud region
        ice_axis = numpy.linspace(ice_bottom, ice_top, n_layer_ice)
        fp.readline() # header line
        if ext:
            n_aux = int(fp.readline().strip().split()[0]) # no. aux

        chan_is_radar = numpy.array(["radar" in x.lower()
                                     for x in chan_names])

        # construct dtype
        #
        # Will have two dtypes:
        # - one where all channels are named
        # - one with a single large dtype for all radiances

        dt = [("IWP", numpy.float32),
              ("Dme", numpy.float32),
              ("Zmed", numpy.float32)]

        dt.extend([("shape", 
                    [("plateagg", numpy.float32),
                     ("sphragg", numpy.float32),
                     ("snowagg", numpy.float32),
                     ("hail", numpy.float32)])])

        if profile:
            dt.append(("IWC", numpy.float32, n_layer_ice))
            dt.append(("Dme_prof", numpy.float32, n_layer_ice))
            dt.append(("RH", numpy.float32, n_layer_rh))

        if ext:
            n_per_aux = n_aux//4
            dt.append(("height", numpy.float32, n_per_aux))
            dt.append(("temp", numpy.float32, n_per_aux))
            dt.append(("pres", numpy.float32, n_per_aux))
            dt.append(("r", numpy.float32, n_per_aux))

        dt_alt = copy.copy(dt)

        dt.extend([(nm, numpy.float32, ln)
                   for (nm, ln) in zip(chan_names, chan_lengths)])

        dt_alt.append(("BT", numpy.float32,
                        numpy.count_nonzero(~chan_is_radar)))
        dt_alt.extend([(nm, numpy.float32, ln)
                       for (nm, ln, israd) in zip(chan_names,
                                                  chan_lengths,
                                                  chan_is_radar)
                       if israd])

#        if radar:
#            dt.extend([(nm, numpy.float32, chanlen)
#                       for (nm, chanlen, israd)
#                       in zip(chan_names, chan_lengths, chan_is_radar)
#                       if israd])

        # get index in measurement array for each measurement
        edges = numpy.r_[0, chan_lengths.cumsum()]
        locations = {chan_names[k]: slice(edges[k], edges[k+1])
                     for k in range(chan_lengths.size)}

#        passive = numpy.array([ln
#                               for ln in chan_lengths
#                               if not "radar" in nm.lower()]

        #n_bt = nchans - sum(["radar" in x.lower() for x in chan_names])

        # pre-allocate
        data = numpy.empty(maxlen, dtype=dt)
        data_alt = data.view(dtype=dt_alt)

        # use while, not for-loop, because I read several lines per
        # iteration
        i = 0
        while i<maxlen:
            line = fp.readline()
            if not line: # empty means EOF
                break
#            if i%nth != 0: # allow for reading less
#                continue
            (rvcheck, IWP, Dme, Zmed,
             plateagg, sphragg, snowagg, hail,
             meltLWP, cldLWP) = (float(x) for x in line.strip().split())
            line = fp.readline()
            RH = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            IWC = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            Dmeprof = numpy.array([float(x) for x in line.strip().split()])
            line = fp.readline()
            measurements = numpy.array([float(x)
                                        for x in line.strip().split()])
            if ext:
                line = fp.readline()
                aux = numpy.array([float(x) for x in
                                   line.strip().split()])
#            BT = numpy.array(measurements[~chan_is_radar])
#            radar = numpy.array(measurements[chan_is_radar])
    #        radar_integrated = numpy.array(measurements[9])
    #        radar_prof = numpy.array(measurements[10:])
            #print "measurements:", BT
            #print "IWP:", IWP
            #print "IWC:", IWC
            data["IWP"][i] = IWP
            data["Dme"][i] = Dme
            data["Zmed"][i] = Zmed
#            data["BT"][i] = BT
            data["shape"][i]["plateagg"] = plateagg
            data["shape"][i]["sphragg"] = sphragg
            data["shape"][i]["snowagg"] = snowagg
            data["shape"][i]["hail"] = hail

            for nm in chan_names:
                data[nm][i] = measurements[locations[nm]]

            if profile:
                data["IWC"][i] = IWC
                data["Dme_prof"][i] = Dmeprof
                data["RH"][i] = RH

            if ext:
                data["height"][i] = aux[0*n_per_aux:1*n_per_aux]
                data["temp"][i] = aux[1*n_per_aux:2*n_per_aux]
                data["pres"][i] = aux[2*n_per_aux:3*n_per_aux]
                data["r"][i] = aux[3*n_per_aux:4*n_per_aux]
#            if radar:
#                for radfield in chan_is_radar.nonzero()[0]:
#                    nm = chan_names[radfield]
#                    data[nm][i] = measurements
            i += 1
            if i%(maxlen//8) == 0:
                print(now(), "done", "{}/{}, {}%".format(i, maxlen,
                                                         (i/maxlen)*100))

        return cls(data=data_alt, instrument=instrument,
                   name="colorado2012", ice_axis=ice_axis)


    @tools.validator
    def __getitem__(self, key: str):
        """Get field from data.

        :param str key: Field name (from self.data["dtype"]).
        :returns: ndarray, view from self.data
        """
        return self.data[key]

    @tools.validator
    def __setitem__(self, key: str, val: numpy.ndarray):
        """Add field to data or set existing field to data.

        :param str key: Field name
        :param ndarray val: Field value.  Must match dimensions of
            self.data.  Note that this will be added with dtype([key,
            val.dtype]).
        """

        if key in self.data.dtype.names:
            self.data[key] = val
        else:
            prim = self.data
            sec = val.view(dtype=[(key, val.dtype)])
            self.data = numpy.lib.recfunctions.merge_arrays(
                (prim, sec)).view(dtype=(prim.dtype.descr + sec.dtype.descr))
            

    # calculating different statistics

#     def stats_IWP_Dme(self, func, minlen=20, *args, **kwargs):
#         """Calculate a statistic per logIWP/Dme
# 
#         :param func: Statistical function to apply for each bin
#         :param minlen: Mask statistic if number of values per bin are less
#             than this.  Defaults to 20.
# 
#         All remaining arguments are passed on to
#         :func:`tools.filter_array`.
# 
#         Returns (xbin, ybin, stats)
#         """
# 
#         xbin = numpy.linspace(-1, 4, 20)
#         ybin = numpy.linspace(0, 650, 20)
#         #pos = self.data["IWP"] > 0
#         minlen = 20
#         nchans = self.data["BT"].shape[1]
#         filler = numpy.empty(nchans)
#         filler.fill(numpy.nan)
#         filt = tools.filter_array(self.data, IWP=(1e-5, 1e10),
#                                   *args, **kwargs)
#         binned_data = tools.bin2d(self.data[filt], "IWP", xbin, "Dme", ybin,
#                                   filter1=numpy.log10)
#         binned_stats = (numpy.array([
#             [func(b["BT"]) if len(b["BT"])>minlen else filler for b in r]
#                 for r in binned_data]))
#         return (xbin, ybin, numpy.ma.masked_invalid(binned_stats))
# 
#     def stats3d(self, func, minlen=20):
#         """Calculate statistic for 3-D bins per IWP, Dme, Zmed
#         """
# 
#         bin_iwp = numpy.linspace(-1, 4, 14)
#         bin_dme = numpy.linspace(0, 650, 15)
#         bin_zmed = numpy.linspace(0, 20, 16)
# 
#         binned = tools.bin3d(self.data,
#            "IWP", bin_iwp,
#            "Dme", bin_dme,
#            "Zmed", bin_zmed,
#            filter1=numpy.log10)
# 
#         nchans = self.data["BT"].shape[1]
#         filler = numpy.empty(nchans)
#         filler.fill(numpy.nan)
# 
#         stats3d = numpy.array([[
#             [func(e["BT"]) if len(e["BT"])>minlen else filler for e in r]
#                 for r in c]
#                     for c in binned])
# 
#         return (bin_iwp, bin_dme, bin_zmed, stats3d)
#  
#     def shape_per_iwp(self):
#         """Get distribution of shapes as a function of IWP
#         """
# 
#         bins_iwp = numpy.linspace(-1, 4, 14)
#         filt = self.data["IWP"] > 0
#         binned = tools.bin(numpy.log10(self.data["IWP"][filt]),
#                            self.data[filt], bins_iwp)
#         shape_dist = numpy.empty_like(bins_iwp, dtype = self.data.dtype["shape"])
#         for shape in self.data.dtype["shape"].names:
#             shape_dist[shape] = [v["shape"][shape].mean() for v in binned]
#         return shape_dist
# 
#     # visualisations
# 
#     def plot_stat_IWP_Dme(self, statfunc, ch,
#                           minlen=20, statname=None, ax=None,
#                           *args, **kwargs):
#         """Visualise statistic IWP Dme in a hist2d.
# 
#         Requires instrument to be defined.
# 
#         Input:
#             - statistic (callable)
#             - channel
#             - minlen, statname, ax
#             - all other passed to tools.filter_array
# 
#         Optional input:
# 
#             - minlen
#             - statname
#             - axes object
# 
#         Returns:
# 
#             fig, ax, pcolormesh, colorbar
#         """
# 
#         if statname is None:
#             statname = statfunc.__name__
# 
#         if ax is None:
#            fig = matplotlib.pyplot.figure()
#            ax = fig.add_subplot(1, 1, 1,
#                xlabel="${}^{10}$log IWP [g/m^2]",
#                xscale="log",
#                ylabel="Dme [µm]",
#                title="{2} {0.name} {0.instrument.name} {1:s} GHz".format(
#                    self,
#                    self.instrument.channels[ch-1].get_chanstr(full=True),
#                    statname))
#        else:
#            fig = ax.figure
#
#        (xbin, ybin, stat) = self.stats_IWP_Dme(statfunc, minlen=minlen,
#                                *args, **kwargs)
#        cmap = matplotlib.cm.Spectral_r
#        cmap.set_bad(color="white", alpha=None)
#        #cmap.set_over(color="black")
#        #cmap.set_under(color="cyan")
#        pm = ax.pcolormesh(10**xbin, ybin, stat[..., ch-1].T, cmap=cmap)
#        #pm.set_clim([0, 15])
#        cb = fig.colorbar(pm)
#        #cb.set_label("interquantile range [K]")
#        
#        return (fig, ax, pm, cb)
#
#    def plot_lines_stat3d(self, statfunc):
#        (bins_iwp, bins_dme, bins_zmed, stat) = self.stats3d(statfunc)
#        isfin = numpy.isfinite(stat[..., 0])
#        fin_in_iwp = isfin.sum(2).sum(1)
#        fin_in_dme = isfin.sum(2).sum(0)
#        fin_in_zmed = isfin.sum(1).sum(0)
#        for (iwp_i, iwp) in enumerate(bins_iwp):
#            if fin_in_iwp[iwp_i] > 10:
#                # one zmed-line per dme for this iwp
#                plot(bins_dme, stat[iwp_i, ..., ch])
#                # one dme-line per zmed for this iwp
#                plot(bins_zmed, stat[iwp_i, ..., ch].T)
##            for (dme_i, dme) in bins_dme:
##                for (zmed_i, zmed) in bins_zmed:
##                    pass
#           

    # writing out the results in different ways

    def write_evans_obs(self, fn, sel=None):
        """Write as evans-obs file

        :param fn: File to write to
        :param sel: Selection to write to file, None (default) to write
            all
        """

        # from iceprofret.txt in the Evans distribution:
        #
        # Six header lines:
        #    Number of pixels
        #    Number of channels, total number of elements
        #    Channel IDs or names (matches channel IDs in retrieval database)
        #    number of elements in each channel
        #    additive uncertainties for each element
        #    multiplicative uncertainties for each element
        # For each line:
        #    Time(hours)  Cosine viewing zenith angle   Measurement for each element


        if sel is None:
            sel = numpy.arange(self.data.size)

        with open(fn, 'wt') as fp:
            # number of pixels
            fp.write("{}\n".format(sel.size))
            # no channels / no elements; both equal to size of BT
            fp.write("{} {}\n".format(self.data.dtype["BT"].shape[0],
                                      self.data.dtype["BT"].shape[0]))
            # names of channels
            fp.write("{}\n".format(self.instrument.channel_string(pre="im",
                                                                  width=True)))

            # no. element per channel
            fp.write(("1 " * len(self.instrument.channels)).strip())
            fp.write("\n")

            # additive uncertainties
            for chan in self.instrument.channels:
                fp.write(str(chan.noise) + " ")
            fp.write("\n")

            # multiplicative uncertainties (not implemented)
            for chan in self.instrument.channels:
                fp.write("0 ")
            fp.write("\n")

            for elem in sel:
                # time; n/a, so write fake
                fp.write("0.0 ")
                # cos(angle); only nadir implented
                fp.write("1.0 ")
                # write channel BT's
                for chan_bt in self.data[elem]["BT"]:
                    fp.write(str(chan_bt) + " ")
                fp.write("\n")


    def write_mat(self, fn, fields=None, sel=None):
        """Write to MAT file.

        Use :func:`scipy.io.savemat` to write the database to a
        MATLAB-style .mat file.
        
        :param fn: Filename to write to
        :param fields: List of fieldnames to write.  Must be subset of
            ``self.data.dtype.names``.  Defaults to ``None``, which means
            to write all fields.
        :param sel: Indices to write.  Defaults to ``None``, which means
            all scenes.
        """
        if fields is None:
            fields = list(self.data.dtype.names)
        if sel is None:
            sel = slice(None)
        print(now(), "Writing to {}".format(fn))
        scipy.io.savemat(fn, dict(data=self.data[fields][sel],
                                  ice_axis=self.ice_axis),
                         appendmat=False,
                         do_compression=True, oned_as="column")
